/*
 * *********************************************************************************************************************************************
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Mario Hellmich <mario.hellmich@web.de> wrote this file. As long as you retain this
 * notice you can do whatever you want with this stuff. If we meet some day,
 * and you think this stuff is worth it, you can buy me a beer in return.
 * Mario Hellmich
 * *********************************************************************************************************************************************
 * File: hw_control.c
 *
 * Version 1.0
 *
 * Created: 21.02.2021
 * Author:  Mario Hellmich <mario.hellmich@web.de>
 * Project: CW Signal Generator Mk-II
 *
 * Low level control and initialization of the synthesizer hardware
 *
 */ 


#include <avr/io.h>

#include "hw_control.h"
#include "globals.h"
#include "system_init.h"

#include <util/delay.h>




/*
************************************************************************************************************************************************
SPI initialization. The hardware uses two wire SPI (transmit only) between the processor and various peripherals, which are addressed by 
dedicated chip select lines.
*/

void spi_init(void)
{
	PORTA.DIRSET = PIN4_bm | PIN6_bm;														// Configure SDA and SCK lines as outputs
	PORTA.DIRCLR = PIN5_bm;																	// Configure (unused) SDI pin as input
	PORTA.DIRCLR = PIN7_bm;																	// Configure (unused) SS pin as input
	
	SPI0.CTRLA |=    SPI_MASTER_bm;															// Configure as master
	SPI0.CTRLA |=    SPI_PRESC_DIV4_gc;														// Set prescaler to divide by 4
	SPI0.CTRLA |=    SPI_CLK2X_bm;															// and enable doubler, i.e., SPI clock frequency is 1.25 MHz
	SPI0.CTRLA &=   ~SPI_DORD_bm;															// Select MSB first transmission as default
	
	SPI0.CTRLB |=    SPI_SSD_bm;															// Disable slave select (SS) pin
	SPI0.CTRLB &=   ~SPI_BUFEN_bm;															// Select normal (unbuffered) mode
	SPI0.CTRLB &= ~( SPI_MODE0_bm | SPI_MODE1_bm );											// Clock line low idle, sample on rising edge (SPI mode 0)
	
	SPI0.CTRLA |=    SPI_ENABLE_bm;															// Enable SPI
}


static void spi_output_reg(uint8_t *addr_ptr, uint8_t len, uint8_t ord)						// Write len bytes from address starting at addr_ptr to the SPI interface with order ord,
{																							// where ord=0: transmit most significant byte first; ord=1: transmit least significant byte first
	if(!len)
		return;
	if(ord)																					// Transmit data LSB first?
	{
		SPI0.CTRLA |= SPI_DORD_bm;															// Transmit byte LSB first
		for( int8_t i = 0; i < len; i++ )
		{
			SPI0.DATA = *(addr_ptr + i);													// Write to SPI data register beginning at addr_ptr (the AVR is little endian)
			while( !(SPI0_INTFLAGS & SPI_IF_bm) )											// Wait until data is exchanged
				;
		}
	}
	else																					// Transmit data MSB first
	{
		SPI0.CTRLA &= ~SPI_DORD_bm;															// Transmit byte MSB first		
		for( int8_t i = len-1; i >= 0; i-- )	
		{
			SPI0.DATA = *(addr_ptr + i);													// Write to SPI data register beginning at addr_ptr + len - 1 (the AVR is little endian)
			while( !(SPI0_INTFLAGS & SPI_IF_bm) )											// Wait until data is exchanged
				;
		}
	}
}




/*
************************************************************************************************************************************************
Port configuration for devices under control
*/

// Define pin number bit masks for hardware connected to the controller
// Chip select, reset and update lines:
#define REFPLL_CS			PIN0_bm
#define MPLL_CS				PIN1_bm
#define	DAC_CS				PIN2_bm

// Harmonic filter:
#define HFILTER0_IN			PIN1_bm
#define HFILTER1_IN			PIN2_bm
#define HFILTER2_IN			PIN3_bm
#define HFILTER0_OUT		PIN4_bm
#define HFILTER1_OUT		PIN5_bm
#define HFILTER2_OUT		PIN6_bm
#define HFILT_BD1_DESEL		PIN7_bm

// Attenuator:
#define ATTEN_D0			PIN0_bm
#define ATTEN_D1			PIN1_bm
#define ATTEN_D2			PIN2_bm
#define ATTEN_D3			PIN3_bm
#define ATTEN_LE			PIN4_bm

// Reference mode selection lines:
#define INT_SEL				PIN5_bm
#define EXT_SEL				PIN6_bm

// USB bridge:
#define USB_RESET			PIN0_bm

// Status indicator input lines
#define REFPLL_LOCK			PIN3_bm
#define MPLL_LOCK			PIN4_bm
#define UNLVLD				PIN5_bm

// Status LEDs
#define STSLED_1			PIN0_bm
#define STSLED_2			PIN1_bm
#define FPLED_RED			PIN2_bm
#define FPLED_GREEN			PIN3_bm


// Define port connections
#define PLL_PORT			PORTB
#define REFPLL_PORT			PORTB
#define DAC_PORT			PORTB
#define HFILTER_PORT		PORTC															// Harmonic filter
#define ATTEN_PORT			PORTD
#define SPI_PORT			PORTB															// Chip select lines
#define REFSEL_PORT			PORTD															// Reference mode selection lines
#define USB_BRIDGE_PORT		PORTC
#define STATUS_PORT			PORTB															// Status indicator lines
#define STSLED_PORT			PORTF															// Status LED lines


// Define chip select and deselect macros for devices connected to the SPI, as well as for the attenuator load enable and the Band 1 select line
#define SELECT_MPLL			PLL_PORT.OUTCLR     = MPLL_CS
#define DESELECT_MPLL		PLL_PORT.OUTSET     = MPLL_CS
#define SELECT_REFPLL		REFPLL_PORT.OUTCLR  = REFPLL_CS
#define DESELECT_REFPLL		REFPLL_PORT.OUTSET  = REFPLL_CS
#define SELECT_DAC			DAC_PORT.OUTCLR     = DAC_CS
#define DESELECT_DAC		DAC_PORT.OUTSET     = DAC_CS
#define SELECT_ATTEN		ATTEN_PORT.OUTSET   = ATTEN_LE
#define DESELECT_ATTEN		ATTEN_PORT.OUTCLR   = ATTEN_LE
#define SELECT_BD1			HFILTER_PORT.OUTCLR = HFILT_BD1_DESEL
#define DESELECT_BD1		HFILTER_PORT.OUTSET = HFILT_BD1_DESEL


static void hw_port_init(void)
{	// Configure pin directions
	// Status indicator lines
	STATUS_PORT.DIRCLR = REFPLL_LOCK | MPLL_LOCK | UNLVLD;									// Configure status indicator lines as input
	
	// USB Bridge
	USB_BRIDGE_PORT.DIRSET = USB_RESET;														// Configure USB bridge reset line as output
	
	// Main PLL
	PLL_PORT.DIRSET = MPLL_CS;																// Configure CS for main PLL as output
	
	// Ref PLL
	REFPLL_PORT.DIRSET = REFPLL_CS;															// Configure CS for reference PLL as output
	
	// DAC
	DAC_PORT.DIRSET = DAC_CS;																// Configure DAC CE as output
	
	// Harmonic filter
	HFILTER_PORT.DIRSET = HFILTER0_IN | HFILTER1_IN | HFILTER2_IN;							// Configure input side harmonic filter control lines as outputs
	HFILTER_PORT.DIRSET = HFILTER0_OUT | HFILTER1_OUT | HFILTER2_OUT;						// Configure output side harmonic filter control lines as outputs
	HFILTER_PORT.DIRSET = HFILT_BD1_DESEL;													// Configure harmonic filter Band 1 deselect line as output 
	
	// Attenuator
	ATTEN_PORT.DIRSET = ATTEN_D0 | ATTEN_D1 | ATTEN_D2 | ATTEN_D3 | ATTEN_LE;				// Configure attenuator control lines as outputs
	
	// Reference mode selection
	REFSEL_PORT.DIRSET = INT_SEL | EXT_SEL;													// Configure reference mode selection lines as outputs
	
	// Status LED outputs
	STSLED_PORT.DIRSET = STSLED_1 | STSLED_2;												// Configure status LED lines as outputs
	
	// Front panel LED outputs
	STSLED_PORT.DIRSET = FPLED_RED | FPLED_GREEN;											// Configure front panel LED lines as output


	// At power up: initial condition and resets
	// De-select all chip enable lines
	DESELECT_MPLL;
	DESELECT_REFPLL;
	DESELECT_DAC;
	DESELECT_ATTEN;
	DESELECT_BD1;

	// Reset USB chip
	USB_BRIDGE_PORT.OUTCLR = USB_RESET;														// Reset is active low
	_delay_ms(20);
	USB_BRIDGE_PORT.OUTSET = USB_RESET;
}




/*
************************************************************************************************************************************************
Stuff for byte conversions
REMEMBER: The AVR is little endian, i.e., var.byte[0] is the least significant byte in var.
*/

typedef union
{
	uint32_t i16;
	uint8_t byte[2];
} convert16to8;

typedef union
{
	uint32_t i32;
	uint8_t byte[4];
} convert32to8;

typedef union
{
	uint32_t i64;
	uint8_t byte[8];
} convert64to8;




/*
************************************************************************************************************************************************
Basic routines to write data to the device registers, functions are call-by-value
*/

static void mpll_write_reg(uint32_t reg)													// Writes a 32 bit value to to the LMX2582 main PLL and VCO circuit
{
	SELECT_MPLL;																			// Chip select line low
	_delay_us(0.1);																			// Wait 100 ns
		convert32to8 buffer;
		buffer.i32 = reg;
		spi_output_reg(buffer.byte, 3, 0);													// Write the first 24 bit of reg MSB first to device 
	_delay_us(0.1);																			// Wait 100 ns
	DESELECT_MPLL;																			// Chip select line high
	_delay_us(0.1);																			// Wait 100 ns (minimum CS high pulse with = 2 ns as per datasheet)
}


static void refpll_write_reg(uint32_t reg)													// Writes the first 24 bit in reg to the ADF4002 reference PLL circuit
{
	SELECT_REFPLL;																			// Chip select line low
	_delay_us(0.1);																			// Wait 100 ns
		convert32to8 buffer;
		buffer.i32 = reg;
		spi_output_reg(buffer.byte, 3, 0);													// Write the first 24 bit of reg MSB first to device
	_delay_us(0.1);																			// Wait 100 ns
	DESELECT_REFPLL;																		// Chip select line high
	_delay_us(0.1);																			// Wait 100 ns (minimum LE high pulse width as per datasheet)
}


static void dac_write_reg(uint16_t reg)														// Writes a 16 bit value to the MCP4921 DAC
{
	SELECT_DAC;																				// Chip select line low
	_delay_us(0.1);																			// Wait 100 ns
		convert16to8 buffer;
		buffer.i16 = reg;
		spi_output_reg(buffer.byte, 2, 0);													// Write 16 bit of reg  MSB first to device
	_delay_us(0.1);																			// Wait 100 ns
	DESELECT_DAC;
	_delay_us(0.1);																			// Wait 100 ns (minimum CS high pulse width as per datasheet)
}




/*
************************************************************************************************************************************************
Write to various hardware registers, e.g. frequency, DAC val, etc., and read hardware status information;
to be accessible from outside this file
*/

void dac_write_val(uint16_t val)															// Write 12 bit value to DAC
{
	uint16_t reg;
	reg = val & 0x0FFF;
	reg |= (1<<13) | (1<<12);																// Set bit 13: output gain times 1, bit 12: no power down 
	dac_write_reg(reg);
}


void refpll_write_mult(uint8_t mult)														// Set reference multiplier, where mult = 100 MHz / (reference frequency)
{
	uint32_t n_counter_latch;
	n_counter_latch   = (1UL<<0)	| (0UL<<1)	|		// Address n counter latch
	(mult << 8)						|					// Set n counter to mult
	(0UL<<21);											// Charge pump current setting 1 is permanently used
	refpll_write_reg(n_counter_latch);
}


void mpll_write_params(struct mpll_s *params)												// Write N divider, numerator and denominator and RF divider settings of main PLL chip
{
	uint32_t reg0, reg35, reg36, reg38, reg40, reg41, reg44, reg45, reg47;
	
	reg0  = (0UL<<16)  |								// Address register 0
			(0UL<<0)   |								// Device power up
			(1UL<<2)   |								// MUX out pin lock detect
			(1UL<<3)   |								// Perform VCO calibration
			(1UL<<4)   |								// Amplitude calibration enable
			(2UL<<7)   |								// FCAL_HPFD_ADJ, used when f_pfd is high, set to f_pfd > 150 MHz
			(0UL<<5)   |								// FCAL_LPFD_ADJ, not used
			(1UL<<9)   |								// Default
			(1UL<<13);									// Lock detect enable
	
	reg35 = (35UL<<16) |								// Address register 35
			(1UL<<0)   |								// Default
			(1UL<<1)   |								// Channel divider segment 1 enable
			( !!(params->chdiv_seg1) << 2 ) |			// Channel divider segment 1 division: chdiv_seg1 = 0: div by 2, chdiv_seg1 = 1: div by 3
			(1UL<<3)   |								// Default
			(1UL<<4)   |								// Default
			(1UL<<7)   |								// Channel divider segment 2 enable
			(1UL<<8)   |								// Channel divider segment 3 enable
			( (params->chdiv_seg2 & 0x0F) << 9 );		// Channel divider segment 2 division
	
	reg36 = (36UL<<16) |								// Address register 36
			( (params->chdiv_seg3 & 0x0F) << 0 )    |	// Channel divider segment 3 division
			( (params->chdiv_seg_sel & 0x07) << 4 ) |	// Channel divider multiplexer
			(1UL<<10);									// Enable buffer between channel divider and output A
	
	reg38 = (38UL<<16) |								// Address register 38
			( (params->pll_int & 0x0FFF) << 1 );		// PLL N divider, 12 bit
	
	reg40 = (40UL<<16) |								// Address register 40
			( (params->pll_den >> 16) & 0xFFFF );		// Denominator of PLL divider fraction, the 16 MSB
	
	reg41 = (41UL<<16) |								// Address register 41
			( params->pll_den & 0xFFFF );				// Denominator of PLL divider fraction, the 16 LSB
	
	reg44 = (44UL<<16) |								// Address register 44
			( (params->pll_num >> 16) & 0xFFFF );		// Numerator of PLL divider fraction, the 16 MSB
	
	reg45 = (45UL<<16) |								// Address register 45
			( params->pll_num & 0xFFFF );				// Numerator of PLL divider fraction, the 16 LSB
	
	reg47 = (47UL<<16) |								// Address register 47
			( !!(params->outa_mux) << 11 ) |			// Route signal to output from VCO or channel divider; outa_mux = 0: from channel divider, outa_mux = 1: from VCO
			(1UL<<6)   |								// Default
			(1UL<<7);									// Default
	
	mpll_write_reg(reg38);																	// Set N divider
	
	mpll_write_reg(reg45);																	// Set numerator of fractional divider
	mpll_write_reg(reg44);
	
	mpll_write_reg(reg41);																	// Set denominator of fractional divider
	mpll_write_reg(reg40);
	
	mpll_write_reg(reg35);																	// Set channel dividers
	mpll_write_reg(reg36);
	mpll_write_reg(reg47);																	// Set channel divider bypass
	
	mpll_write_reg(reg0);																	// Write register 0 last to start VCO calibration
}


void set_ref(uint8_t mode)																	// Switch between internal and external reference, mode = 0: internal, mode = 1: external
{
	if(mode)																				// Use external reference?
	{
		REFSEL_PORT.OUTSET = EXT_SEL;
		REFSEL_PORT.OUTCLR = INT_SEL;
		refpll_write_mult(ref_mult);														// Set reference multiplier to value of corresponding global variable
	}
	else																					// Use internal reference
	{
		REFSEL_PORT.OUTSET = INT_SEL;
		REFSEL_PORT.OUTCLR = EXT_SEL;
		refpll_write_mult( (uint8_t)( mpll_fref / f_int_ref ) );							// Set reference multiplier to match internal reference
	}
}


void set_atten(uint8_t att_val)																// Select output attenuator value
{
	DESELECT_ATTEN;
	if( att_val > 15 )																		// Maximum attenuator value = 2^4 - 1 = 15
		return;	
	ATTEN_PORT.OUTCLR = ATTEN_D0 | ATTEN_D1 | ATTEN_D2 | ATTEN_D3;							// Clear bits in GPIO output register
	ATTEN_PORT.OUTSET = (ATTEN_D0 | ATTEN_D1 | ATTEN_D2 | ATTEN_D3) & ~att_val;				// Set GPIO register
	_delay_us(0.02);
	SELECT_ATTEN;																			// Attenuator latch enable line high for 50 ns
	_delay_us(0.05);
	DESELECT_ATTEN;
}


void set_hfilter(uint8_t val)																// Set harmonic filter; val = 0...8: filter numbers, corresponds to Band 1 ... Band 9
{
	if( val > 8 )																			// Check for argument out of range
		return;
	HFILTER_PORT.OUTCLR = (HFILTER0_IN | HFILTER1_IN | HFILTER2_IN);						// Clear bits in GPIO output register
	HFILTER_PORT.OUTCLR = (HFILTER0_OUT | HFILTER1_OUT | HFILTER2_OUT);
	if( val <= 1 )																			// Band 1 and 2 (filter 0 and 1) are switched by the two SPDT switches U9 and U10
	{
		HFILTER_PORT.OUTSET = (HFILTER0_IN | HFILTER1_IN | HFILTER2_IN) & ( 7 << 1 );		// Select output RF8 on U13
		HFILTER_PORT.OUTSET = (HFILTER0_OUT | HFILTER1_OUT | HFILTER2_OUT) & ( 0 << 4 );	// Select output RF1 on U14
		if( val == 0 )																		// Is Band 1 selected?
			SELECT_BD1;																		// Then set U9 and U10 for Band 1
		else																				// Band 2 is selected
			DESELECT_BD1;																	// Set U9 and U10 for Band 2
	}
	else																					// Band 3...9 (filter 2...8) are selected
	{
		val--;																				// Decrement val by one because filters 2...8 are connected to U13 and U14 directly
		HFILTER_PORT.OUTSET = (HFILTER0_IN | HFILTER1_IN | HFILTER2_IN) & ( (7-val) << 1 );	// Set U13
		HFILTER_PORT.OUTSET = (HFILTER0_OUT | HFILTER1_OUT | HFILTER2_OUT) & ( val << 4 );	// Set U14
	}
}


void set_stsled_1(uint8_t led_on)															// Switch status LED 1
{
	if(led_on)
		STSLED_PORT.OUTSET = STSLED_1;
	else
		STSLED_PORT.OUTCLR = STSLED_1;
}


void set_stsled_2(uint8_t led_on)															// Switch status LED 2
{
	if(led_on)
		STSLED_PORT.OUTSET = STSLED_2;
	else
		STSLED_PORT.OUTCLR = STSLED_2;
}


void toggle_stsled_2(void)																	// Toggle status LED 2
{
	STSLED_PORT.OUTTGL = STSLED_2;
}


void set_fpled_err(uint8_t err_status)														// Set front panel indicator LED, led_status = 0: green, led_status = 1: red
{
	if(!err_status)
	{
		STSLED_PORT.OUTSET = FPLED_GREEN;
		STSLED_PORT.OUTCLR = FPLED_RED;	
	}
	else
	{
		STSLED_PORT.OUTCLR = FPLED_GREEN;
		STSLED_PORT.OUTSET = FPLED_RED;
	}
}


uint8_t get_refpll_unlk(void)																// Get reference PLL lock status
{
	if(STATUS_PORT.IN & REFPLL_LOCK)
		return FALSE;
	else
		return TRUE;
}


uint8_t get_mpll_unlk(void)																	// Get main PLL lock status
{
	if(STATUS_PORT.IN & MPLL_LOCK)
		return FALSE;
	else
		return TRUE;
}


uint8_t get_alc_unlvld(void)																// Get ALC status
{
	if(STATUS_PORT.IN & UNLVLD)
		return TRUE;
	else
		return FALSE;
}




/*
************************************************************************************************************************************************
Port and device initialization
*/

static void refpll_init(void)
{
	uint32_t ref_counter_latch, n_counter_latch, funct_latch, init_latch;
	
	ref_counter_latch = (0UL<<0)  | (0UL<<1)      |		// Address reference counter latch
						(1UL<<2)  |						// Set reference divider to 1
						(0UL<<16) | (0UL<<17)     |		// 2.9ns anti-backslash pulse width
						(0UL<<20) ;						// Fast lock detect
	
	n_counter_latch   = (1UL<<0)  | (0UL<<1)      |		// Address n counter latch
						(0x000A << 8)             |		// Set n counter to 10 (100 MHz VCO frequency)
						(0UL<<21) ;						// Charge pump current setting 1 is permanently used
	
	funct_latch       = (0UL<<0)  | (1UL<<1)      |		// Address function latch
						(0UL<<2)  |						// Counter normal operation
						(0UL<<21) | (0UL<<3)      |		// No power-down, normal operation
						(0UL<<6)|(0UL<<5)|(1UL<<4)|		// MUXOUT digital lock detect
						(1UL<<7)  |						// PFD polarity positive
						(0UL<<8)  |						// CP not in three state
						(0UL<<9)  | (0UL<<10)     |		// Fast lock disabled
						(0x0007 << 11)            |		// Charge pump current setting 1 (5 mA)
						(0x0007 << 14) ;       			// Charge pump current setting 2 (5 mA)
	
	init_latch		  = (0UL<<1)  | (1UL<<1)      |		// Address initialization latch
						(0UL<<2)  |						// Counter normal operation
						(0UL<<21) | (0UL<<3)      |		// No power-down, normal operation
						(0UL<<6)|(0UL<<5)|(1UL<<4)|		// MUXOUT digital lock detect
						(1UL<<7)  |						// PFD polarity positive
						(0UL<<8)  |						// CP not in three state
						(0UL<<9)  | (0UL<<10)     |		// Fast lock disabled
						(0x0007 << 11)            |		// Charge pump current setting 1 (5 mA)
						(0x0007 << 14) ;      			// Charge pump current setting 2 (5 mA)
	
	refpll_write_reg(init_latch);
	refpll_write_reg(funct_latch);
	refpll_write_reg(ref_counter_latch);
	refpll_write_reg(n_counter_latch);
}


static void mpll_init(void)
{
	uint32_t reg0, reg1, reg2, reg4, reg7, reg8, reg9, reg10, reg11, reg12, reg13, reg14, reg19, reg20, reg22, reg23, reg24, reg25, reg28, reg29, reg30, reg31, reg32, reg33, reg34, reg37, reg39, reg42, reg43, reg46, reg48, reg59, reg61, reg62, reg64, reg68, reg69, reg70;
		
	reg0  = (0UL<<16)  |								// Address register 0
			(1UL<<1)   |								// Perform soft reset
			(1UL<<9);									// Default
	mpll_write_reg(reg0);								// Write register 0 to perform a soft reset
	_delay_ms(250);										// Wait 250 ms
	
	reg0  = (0UL<<16)  |								// Address register 0
			(0UL<<0)   |								// Device power up
			(1UL<<2)   |								// MUX out pin lock detect
			(1UL<<3)   |								// Perform VCO calibration
			(1UL<<4)   |								// Amplitude calibration enable
			(0UL<<7)   |								// FCAL_HPFD_ADJ, used when f_pfd is high, set to f_pfd <= 150 MHz
			(0UL<<5)   |								// FCAL_LPFD_ADJ, not used
			(1UL<<9)   |								// Default
			(1UL<<13);									// Lock detect enable
	
	reg1  = (1UL<<16)  |								// Address register 1
			(0UL<<0)   |								// Set CAL_CLK_DIV to 2^0 = 1
			(1UL<<3)   |								// Default
			(1UL<<11);									// Default
	
	reg2  = 0x020500;									// Register as per TICSpro
	
	reg4  = (4UL<<16)  |								// Address register 4
			( (25 & 0xFF) << 8 ) |						// Set VCO amplitude calibration delay to 25
			(1UL<<0)   |								// Default
			(1UL<<1)   |								// Default
			(1UL<<6);									// Default
			
	reg7 = 0x0728B2;									// Register as per TICSpro
	
	reg8  = (8UL<<16)  |								// Address register 8
			(1UL<<2)   |								// Default
			(1UL<<7)   |								// Default
			(0UL<<10)  |								// Override VCO capacitor bank selection disable
			(1UL<<12)  |								// Default
			(0UL<<13);									// Override VCO amplitude disable
	
	reg9  = (9UL<<16)  |								// Address register 9
			(1UL<<1)   |								// Default
			(1UL<<8)   |								// Default
			(1UL<<9)   |								// Enable reference path
			(0UL<<11);									// Disable reference path doubler
	
	reg10 = (10UL<<16) |								// Address register 10
			(1UL<<3)   |								// Default
			(1UL<<4)   |								// Default
			(1UL<<6)   |								// Default
			( (1 & 0x1F) << 7 ) |						// Set reference multiplier to 1
			(1UL<<12);									// Default
	
	reg11 = (11UL<<16) |								// Address register 11
			(1UL<<3)   |								// Default
			( (1 & 0xFF) << 4 );						// Set R-divider after multiplier and before PFD to 1
	
	reg12 = (12UL<<16) |								// Address register 12
			( (1UL & 0x0FFF) ) |						// Set R-divider after OSC-in doubler and before multiplier to 1
			(1UL<<12)  |								// Default
			(1UL<<13)  |								// Default
			(1UL<<14);									// Default
	
	reg13 = (13UL<<16) |								// Address register 13
			(1UL<<14)  |								// Enable charge pump
			(0UL<<8);									// Select dual CP mode
	
	reg14 = (14UL<<16) |								// Address register 14
			(3UL<<0)   |								// Set charge pump gain multiplier to times 2.5
			(0x1F<<2)  |								// Set charge pump up current to maximum
			(0x1F<<7);									// Set charge pump down current to maximum
	
	reg19 = 0x130965;									// Register as per TICSpro
	reg20 = 0x14012C;									// Register as per TICSpro
	reg22 = 0x162300;									// Register as per TICSpro
	reg23 = 0x178842;									// Register as per TICSpro
	reg24 = 0x180509;									// Register as per TICSpro
	reg25 = 0x190000;									// Register as per TICSpro
	reg28 = 0x1C2924;									// Register as per TICSpro
	reg29 = 0x1D0084;									// Register as per TICSpro
	
	reg30 = (30UL<<16) |								// Address register 30
			(1UL<<2)   |								// Default
			(1UL<<4)   |								// Default
			(1UL<<5)   |								// Default
			(1UL<<10);									// Select MASH dithering on
	
	reg31 = (31UL<<16) |								// Address register 31
			(1UL<<0)   |								// Default
			(0UL<<7)   |								// Power up buffer between VCO and channel divider
			(0UL<<9)   |								// Power up buffer between VCO and output A
			(1UL<<10);									// Power down buffer between VCO and output B
	
	reg32 = 0x20210A;									// Register as per TICSpro
	reg33 = 0x212A0A;									// Register as per TICSpro
	
	reg34 = (34UL<<16) |								// Address register 31
			(1UL<<1)   |								// Default
			(1UL<<3)   |								// Default
			(1UL<<5)   |								// Enable channel divider
			(1UL<<6)   |								// Default
			(1UL<<7)   |								// Default
			(1UL<<8)   |								// Default
			(1UL<<9)   |								// Default
			(1UL<<14)  |								// Default
			(1UL<<15);									// Default
	
	reg37 = (37UL<<16) |								// Address register 37
			(0UL<<12)  |								// Set N divider prescaler, prescaler = 0: divide-by-2, prescaler = 1: divide-by-4
			(1UL<<14);									// Default
	
	reg39 = 0x278204;									// Register as per TICSpro
	reg42 = 0x2A0000;									// Register as per TICSpro
	reg43 = 0x2B0000;									// Register as per TICSpro
	
	reg46 = (46UL<<16) |								// Address register 46
			(1UL<<0)   | (1UL<<1) | (0UL<<2) |			// Set MASH order to 3
			(1UL<<5)   |								// Default
			(0UL<<6)   |								// Power up output A
			(1UL<<7)   |								// Power down output B
			( (63 & 0x3F) << 8 );						// Set output A power to 63 (min: 0, max: 63)
	
	reg48 = 0x3003FC;									// Register as per TICSpro
	
	reg59 = (59UL<<16) |								// Address register 59
			(1UL<<5);									// Select high current drive on MUXOUT pin
	
	reg61 = (61UL<<16) |								// Address register 61
			(1UL<<0);									// Lock detect type, set to tuning voltage range
	
	reg62 = 0x3E0000;									// Register as per TICSpro
	reg64 = 0x400077;									// Register as per TICSpro
	reg68 = 0x440089;									// Register as per TICSpro
	reg69 = 0x450000;									// Register as per TICSpro
	reg70 = 0x460000;									// Register as per TICSpro
	
	
	mpll_write_reg(reg1);								// Write registers to chip
	mpll_write_reg(reg2);
	mpll_write_reg(reg4);
	mpll_write_reg(reg7);
	mpll_write_reg(reg8);
	mpll_write_reg(reg9);
	mpll_write_reg(reg10);
	mpll_write_reg(reg11);
	mpll_write_reg(reg12);
	mpll_write_reg(reg13);
	mpll_write_reg(reg14);
	mpll_write_reg(reg19);
	mpll_write_reg(reg20);
	mpll_write_reg(reg22);
	mpll_write_reg(reg23);
	mpll_write_reg(reg24);
	mpll_write_reg(reg25);
	mpll_write_reg(reg28);
	mpll_write_reg(reg29);
	mpll_write_reg(reg30);
	mpll_write_reg(reg31);
	mpll_write_reg(reg32);
	mpll_write_reg(reg33);
	mpll_write_reg(reg34);
	mpll_write_reg(reg37);
	mpll_write_reg(reg39);
	mpll_write_reg(reg42);
	mpll_write_reg(reg43);
	mpll_write_reg(reg46);
	mpll_write_reg(reg48);
	mpll_write_reg(reg59);
	mpll_write_reg(reg61);
	mpll_write_reg(reg62);
	mpll_write_reg(reg64);
	mpll_write_reg(reg68);
	mpll_write_reg(reg69);
	mpll_write_reg(reg70);
	
	mpll_write_reg(reg0);								// Write register 0 last to start amplitude and frequency calibration
}


void hw_init(void)																			// To be called in main once at system startup
{
	hw_port_init();
	refpll_init();																			// Initialize reference PLL
	_delay_ms(250);																			// Wait 250 ms for main PLL reference to settle
	mpll_init();																			// Initialize main PLL
}